from sortedcontainers import SortedKeyList
from abc import abstractmethod, ABC
from .l2_book_side import L2BookSide, SortedKeyList


class AbstractL2Book(ABC):
    """

    """

    def __init__(self, instrument: str):
        """

        :param instrument:
        """
        self.instrument = instrument
        self._bids = L2BookSide(is_bids=True)
        self._asks = L2BookSide(is_bids=False)

    @abstractmethod
    def handle_snapshot(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        pass

    @abstractmethod
    def handle_update(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        pass

    @property
    def bids(self):
        return self._bids.data

    @property
    def asks(self):
        return self._asks.data



import re
from typing import Pattern


def resolve_route(value, routes):
    key, handler = None, None
    for key, handler in routes:
        if key:
            if isinstance(key, str):
                if key == value:
                    return handler
            elif isinstance(key, Pattern) and re.match(key, value):
                return handler

    if key is not None and key == "" and handler:
        return handler


def hide_secret(request):
    data = request.copy()
    hidden_keys = ["password", "username", "client_id", "client_secret", "refresh_token", "access_token"]

    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = hide_secret(value)
        elif key in hidden_keys:
            data[key] = "***"

    return data


class IntId:
    id = 0

    def get_id(self):
        self.id += 1
        return self.id



import asyncio
import logging

import aiohttp

logger = logging.getLogger(__name__)


class SessionWrapper:
    __internal_session: bool = True
    __is_session: bool = False
    __session: aiohttp.ClientSession = None
    _timeout: aiohttp.ClientTimeout = None

    def __init__(self, timeout: aiohttp.ClientTimeout = None, session: aiohttp.ClientSession = None):
        if session:
            self._session = session
            self.__internal_session = False

        if timeout:
            self._timeout = timeout
        else:
            self._timeout = aiohttp.ClientTimeout(total=20)

    @property
    def _session(self):
        if not self.__is_session:
            self.__check_session()
        return self.__session

    @_session.setter
    def _session(self, value):
        self.__is_session = True
        self.__internal_session = False
        self.__session = value

    def __check_session(self):
        if not self.__is_session:
            self.__is_session = True
            self.__internal_session = True
            self.__session = aiohttp.ClientSession(timeout=self._timeout)

    def __del__(self):
        self._close()

    def __enter__(self):
        self.__check_session()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._close()

    def _close(self):
        if self.__internal_session and self.__session:
            asyncio.ensure_future(self.__session.close())
            self.__is_session = False



import logging
from time import time

import aiohttp

from ssc2ce.common.session import SessionWrapper
from ssc2ce.common.utils import resolve_route

from enum import IntEnum


class Bitfinex(SessionWrapper):
    class ConfigFlag(IntEnum):
        TIMESTAMP = 32768
        SEQ_ALL = 65536
        CHECKSUM = 131072

    class StatusFlag(IntEnum):
        MAINTENANCE = 0
        OPERATIVE = 1

    ws: aiohttp.ClientWebSocketResponse = None
    on_connect_ws = None
    on_close_ws = None
    on_maintenance = None
    on_conf = None
    receipt_time = None

    ws_api = 'wss://api-pub.bitfinex.com/ws/2'

    last_message = None
    is_connected = False
    subscriptions = []
    channel_handlers = {}

    def __init__(self,
                 flags: ConfigFlag = ConfigFlag.TIMESTAMP | ConfigFlag.SEQ_ALL):
        super().__init__()

        self.flags = flags
        self.logger = logging.getLogger(__name__)

        self._timeout: aiohttp.ClientTimeout = aiohttp.ClientTimeout(total=20)

        self.on_message = self.handle_message
        # self.on_connect = self.configure
        self.routes = [
            ("subscribed", self.handle_subscribed),
            ("info", self.handle_info),
            ("conf", self.handle_conf),
            # ("", self.warn_handler),
        ]

        # self.channel_route = []

    async def configure(self, flags: ConfigFlag = None):
        if flags is not None:
            self.flags = flags

        if self.flags is not None:
            request = dict(event="conf", flags=self.flags)
            await self.ws.send_json(request)

    async def subscribe(self, request, handler):
        self.subscriptions.append((request, handler))
        self.logger.info(f"subscribe {request}")
        await self.ws.send_json({
            "event": "subscribe",
            **request
        })

    async def run_receiver(self):
        self.ws = await self._session.ws_connect(self.ws_api)

        while self.ws and not self.ws.closed:
            message = await self.ws.receive()
            self.receipt_time = time()
            self.last_message = message

            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.ERROR):
                self.logger.warning(f"Connection close {repr(message)}")
                if self.on_close_ws:
                    await self.on_close_ws()

                continue
            if message.type == aiohttp.WSMsgType.CLOSING:
                self.logger.debug(f"Connection closing {repr(message)}")
                continue

            if self.on_message:
                await self.on_message(message)

    async def handle_message(self, message: aiohttp.WSMessage):
        if message.type == aiohttp.WSMsgType.TEXT:
            data = message.json()

            if isinstance(data, list):
                channel_id = data[0]
                handler = self.channel_handlers.get(channel_id)
                if handler:
                    await handler(data, self)
                else:
                    self.logger.warning(f"Can't find handler for channel_id{channel_id}, {data}")
            elif isinstance(data, dict):
                if "event" in data:
                    await self.handle_event(data)
                else:
                    self.logger.warning(f"Unknown message {message.data}")
            else:
                self.logger.warning(f"Unknown message {message.data}")
        else:
            self.logger.warning(f"Unknown type of message {repr(message)}")

    async def empty_handler(self, data):
        pass

    async def handle_subscribed(self, message):
        self.logger.info(f"receive subscribed: {message}")
        idx = None
        for i, s in enumerate(self.subscriptions):
            if s[0].items() <= message.items():
                idx = i
                self.channel_handlers[message["chanId"]] = s[1]

        if idx:
            del self.subscriptions[idx]

    async def handle_conf(self, message):
        """{'event': 'conf', 'status': 'OK', 'flags': 98304}"""
        self.logger.info(f"{message}")
        if self.on_conf:
            await self.on_conf()

    async def handle_info(self, message):
        """
        Handle an info message that contains the actual version of the websocket stream, along with a platform status
        flag (1 for operative, 0 for maintenance
        :param message: Info messages {'event': 'info', 'version': 2, 'serverId': ..., 'platform': {'status': 1}}
          version: the actual version of the websocket stream must be 2
          status: a platform status flag (1 for operative, 0 for maintenance).
        :return:
        """
        self.logger.info(f"{message}")

        if message["version"] != 2:
            raise NotImplemented(f"Bitfinex connector support only version 2 but receive {message}")

        if message["platform"]["status"] == 1:
            if not self.is_connected:
                await self.configure()
                if self.on_connect_ws:
                    await self.on_connect_ws()
            else:
                if self.on_maintenance:
                    await self.on_maintenance(message)
        else:
            if self.on_maintenance:
                await self.on_maintenance(message)

    async def warn_handler(self, message):
        self.logger.warning(f"Unsupported message {message}")

    async def handle_event(self, message):
        event = message["event"]
        handler = resolve_route(event, self.routes)

        if handler:
            return await handler(message)

        self.logger.warning(f"Unhandled event:{event}, message:{repr(message)} .")
        return

    def close(self):
        super()._close()

    async def stop(self):
        await self.ws.close()
        self.close()



from .deribit import Deribit
from .common import AuthType
from .bitfinex import Bitfinex

from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass


#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='ssc2ce',
    version="0.10.0",
    author='Oleg Nedbaylo',
    author_email='olned64@gmail.com',
    description='A Set of Simple Connectors for access To Cryptocurrency Exchanges',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/olned/ssc2ce-python',
    packages=setuptools.find_packages(),
    install_requires=['aiohttp'],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import logging
from ssc2ce.common.abstract_l2_book import AbstractL2Book
from collections import deque

from ssc2ce.common.exceptions import BrokenOrderBook


class L2Book(AbstractL2Book):
    change_id = None
    timestamp = None
    logger = logging.getLogger(__name__)

    def __init__(self, instrument: str):
        """

        :param instrument:
        """
        AbstractL2Book.__init__(self, instrument)

    def handle_snapshot(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        self.asks.clear()
        self.bids.clear()

        self.change_id = message["change_id"]
        self.timestamp = message["timestamp"]
        for i in message['bids']:
            self.bids.add([i[1], i[2]])

        for i in message['asks']:
            self.asks.add([i[1], i[2]])

    def handle_update(self, message: dict) -> None:
        """

        :param message:
        :return:
        """
        prev_change_id = message["prev_change_id"]
        if prev_change_id != self.change_id:
            raise BrokenOrderBook(self.instrument, prev_change_id, self.change_id)

        self.change_id = message["change_id"]
        self.timestamp = message["timestamp"]
        for change in message['bids']:
            if change[0] == 'new':
                self._bids.add(change[1:])
            elif change[0] == 'delete':
                self._bids.delete(change[1])
            else:
                self._bids.update(price=change[1], size=change[2])

        for change in message['asks']:
            if change[0] == 'new':
                self._asks.add(change[1:])
            elif change[0] == 'delete':
                self._asks.delete(change[1])
            else:
                self._asks.update(price=change[1], size=change[2])


def create_l2_order_book(instrument: str) -> AbstractL2Book:
    return L2Book(instrument)
